#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/7/22 14:29
# @Author  : zhangbc0315@outlook.com
# @File    : __init__.py.py
# @Software: PyCharm

from chem_utils.chem_handler.atom_handler import AtomHandler
from chem_utils.chem_handler.mol_handler import MolHandler


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/8/15 17:15
# @Author  : zbc@mail.ustc.edu.cn
# @File    : rxn_smiles.py
# @Software: PyCharm


from rdkit.Chem import AllChem
from rdkit.Chem.AllChem import ChemicalReaction


class RxnSmiles:

    # region ===== rdrxn_from_indigo_smiles =====

    @classmethod
    def _parse_indigo_tag(cls, indigo_tags_str: str) -> [[int]]:

        for tags in indigo_tags_str.split(','):
            pass
        indigo_tags = [[int(tag) for tag in tags.split('.')] for tags in indigo_tags_str.split(',')]
        return indigo_tags

    @classmethod
    def _split_smiles_by_indigo_tags(cls, rxn_smiles: str, indigo_tags: [[int]]) -> ([str], [str]):
        rs_smiles, ps_smiles = rxn_smiles.split('>>')
        r_smileses = rs_smiles.split('.')
        p_smileses = ps_smiles.split('.')
        all_smileses = r_smileses.copy()
        all_smileses.extend(p_smileses)
        num_rs = len(r_smileses)
        rs = []
        ps = []
        all_tags = []
        for tags in indigo_tags:
            all_tags.extend(tags)
            mol_smileses = []

            for tag in tags:
                mol_smileses.append(all_smileses[tag])
            if tags[0] >= num_rs:
                ps.append('.'.join(mol_smileses))
            else:
                rs.append('.'.join(mol_smileses))
        for i, smiles in enumerate(all_smileses):
            if i in all_tags:
                continue
            if i >= num_rs:
                ps.append(smiles)
            else:
                rs.append(smiles)
        return rs, ps

    @classmethod
    def _rdrxn_from_smileses(cls, r_smileses: [str], p_smileses: [str]):
        rdrxn = ChemicalReaction()
        for r_smiles in r_smileses:
            rdrxn.AddReactantTemplate(AllChem.MolFromSmiles(r_smiles))
        for p_smiles in p_smileses:
            rdrxn.AddProductTemplate(AllChem.MolFromSmiles(p_smiles))
        return rdrxn

    @classmethod
    def _split_indigo_smiles(cls, indigo_smiles: str) -> (str, str):
        if ' |f:' in indigo_smiles:
            rxn_smiles, indigo_tags_str = indigo_smiles.split(' |f:')
        else:
            rxn_smiles = indigo_smiles
            indigo_tags_str = ''

        if '^' in indigo_tags_str:
            indigo_tags_str = indigo_tags_str.split('^')[0]
        indigo_tags_str = indigo_tags_str.strip('|')
        return rxn_smiles, indigo_tags_str

    @classmethod
    def rdrxn_from_indigo_smiles(cls, indigo_smiles: str):
        rxn_smiles, indigo_tags_str = cls._split_indigo_smiles(indigo_smiles)
        indigo_tags = cls._parse_indigo_tag(indigo_tags_str)
        r_smileses, p_smileses = cls._split_smiles_by_indigo_tags(rxn_smiles, indigo_tags)
        return cls._rdrxn_from_smileses(r_smileses, p_smileses)

    # endregion


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/8/15 17:17
# @Author  : zbc@mail.ustc.edu.cn
# @File    : rxn_cleaner.py
# @Software: PyCharm

from typing import Optional

from rdkit.Chem import AllChem
from rdkit.Chem.rdchem import Mol


class RXNCleaner:

    @classmethod
    def _is_mol_mapped(cls, mol: Mol) -> bool:
        for atom in mol.GetAtoms():
            if atom.GetAtomMapNum() != 0:
                return True
        return False

    @classmethod
    def _get_mapped_mols(cls, mols: [Mol]) -> [Mol]:
        new_mols = []
        for mol in mols:
            if cls._is_mol_mapped(mol):
                new_mols.append(mol)
        return new_mols

    @classmethod
    def clean_unmapped(cls, rdrxn: AllChem.ChemicalReaction) -> Optional[AllChem.ChemicalReaction]:
        products = rdrxn.GetProducts()
        reactants = rdrxn.GetReactants()

        mapped_reactants = cls._get_mapped_mols(reactants)
        mapped_products = cls._get_mapped_mols(products)
        if len(mapped_products) == 0 or len(mapped_reactants) == 0:
            return None
        new_rdrxn = AllChem.ChemicalReaction()
        for reactant in mapped_reactants:
            new_rdrxn.AddReactantTemplate(reactant)
        for product in mapped_products:
            new_rdrxn.AddProductTemplate(product)
        return new_rdrxn

    @classmethod
    def is_mappable(cls, rdrxn: AllChem.ChemicalReaction) -> bool:
        products = rdrxn.GetProducts()
        reactants = rdrxn.GetReactants()
        reactants_map_nums = set()
        for reactant in reactants:
            for atom in reactant.GetAtoms():
                reactants_map_nums.add(atom.GetAtomMapNum())
        for product in products:
            for atom in product.GetAtoms():
                map_num = atom.GetAtomMapNum()
                if map_num == 0:
                    return False
                elif map_num not in reactants_map_nums:
                    return False
        return True


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/7/22 14:36
# @Author  : zhangbc0315@outlook.com
# @File    : mol_handler.py
# @Software: PyCharm
import logging
from typing import Iterator, Tuple, Dict

from rdkit.Chem.rdchem import Mol, EditableMol, Atom, BondType
from rdkit.Chem import AllChem


class MolHandler:

    # region ===== correct mol =====

    @classmethod
    def link_to_atom_by_bond(cls, atom, atomic_num: int, bond_type):
        for bond in atom.GetBonds():
            if bond.GetBondType() != bond_type:
                continue
            atom1 = bond.GetBeginAtom()
            atom2 = bond.GetEndAtom()
            if atom1 == atom and atom2.GetAtomicNum() == atomic_num:
                return True, bond, atom2
            elif atom2 == atom and atom1.GetAtomicNum() == atomic_num:
                return True, bond, atom1
        return False, None, None

    @classmethod
    def count_bond_weight(cls, atom):
        res = 0
        for bond in atom.GetBonds():
            bond_type_level = int(bond.GetBondType())
            if bond_type_level <= 3:
                res += bond_type_level
            else:
                res += 1.5
        return res

    @classmethod
    def correct_n(cls, n_atom, rdmol):
        bond_weight = cls.count_bond_weight(n_atom)
        if n_atom.GetTotalNumHs() == 0:
            return
        if bond_weight + n_atom.GetTotalNumHs() < 4:
            return
        n_atom.SetNumExplicitHs(0)
        mol = AllChem.MolFromSmiles(AllChem.MolToSmiles(rdmol))
        if mol is not None:
            return
        n_atom.SetNumExplicitHs(1)

    @classmethod
    def correct_mol(cls, rdmol):
        mol = AllChem.MolFromSmiles(AllChem.MolToSmiles(rdmol))
        if mol is not None:
            return
        # print(AllChem.MolToSmiles(rdmol))
        for atom in rdmol.GetAtoms():
            if atom.GetAtomicNum() == 7:
                try:
                    cls.correct_n(atom, rdmol)
                except:
                    continue
        # print(AllChem.MolToSmiles(rdmol))

    # endregion

    @classmethod
    def get_inchis_from_smileses(cls, smileses: [str]):
        inchis = []
        for smiles in smileses:
            mol = AllChem.MolFromSmiles(smiles)
            if mol is None:
                return None
            inchis.append(AllChem.MolToInchi(mol))
        return inchis

    @classmethod
    def remove_inchi_stereo(cls, inchi: str):
        mol = AllChem.MolFromInchi(inchi)
        AllChem.RemoveStereochemistry(mol)
        return AllChem.MolToInchi(mol)

    @classmethod
    def atom_num(cls, smiles):
        mol = AllChem.MolFromSmiles(smiles)
        return mol.GetNumAtoms()

    @classmethod
    def bond_num(cls, smiles):
        mol = AllChem.MolFromSmiles(smiles)
        if mol is None:
            logging.warning(f"wrong smiles: {smiles}")
            return 0
        return mol.GetNumBonds()

    @classmethod
    def remove_charge(cls, rdmol):
        for atom in rdmol.GetAtoms():
            atom.SetFormalCharge(0)

    @classmethod
    def contain_smiles_or_error(cls, smileses: [str], smiles: str):
        if smiles in smileses:
            return True
        try:
            inchis = [AllChem.MolToInchi(AllChem.MolFromSmiles(smi)) for smi in smileses]
        except:
            return True
        if any([inc is None for inc in inchis]):
            return True
        inchi = AllChem.MolToInchi(AllChem.MolFromSmiles(smiles))
        if inchi is None:
            return True
        return inchi in inchis

    # region ===== smiles inchi =====

    @classmethod
    def smiles_from_inchi(cls, inchi: str):
        return AllChem.MolToSmiles(AllChem.MolFromInchi(inchi))

    @classmethod
    def inchi_from_smiles(cls, smiles: str):
        return AllChem.MolToInchi(AllChem.MolFromSmiles(smiles))

    # endregion

    # region ===== smiles/inchi to mol =====
    @classmethod
    def mol_from_smiles(cls, smiles: str):
        try:
            mol = AllChem.MolFromSmiles(smiles)
        except:
            mol = None
        return mol

    @classmethod
    def mol_from_inchi(cls, inchi: str):
        try:
            mol = AllChem.MolFromInchi(inchi)
            smi = AllChem.MolToSmiles(mol)
            mol = AllChem.MolFromSmiles(smi)
        except:
            mol = None
        return mol

    @classmethod
    def mol_from_inchis(cls, inchis: [str]):
        mol = None
        for inchi in inchis:
            if mol is None:
                mol = AllChem.MolFromInchi(inchi)
            else:
                mol = AllChem.CombineMols(mol, AllChem.MolFromInchi(inchi))
        return mol

    # region ===== get mol fingerprint =====

    @classmethod
    def get_mol_fp_by_rdmol(cls, rdmol: Mol):
        return AllChem.GetMorganFingerprintAsBitVect(rdmol)

    @classmethod
    def get_mol_fp_by_smiles(cls, smiles: str):
        return cls.get_mol_fp_by_rdmol(AllChem.MolFromSmiles(smiles))

    # endregion

    # region ===== get smiles =====

    @classmethod
    def get_smiles_without_map_num(cls, mol: Mol) -> str:
        copy_mol = AllChem.MolFromSmiles(AllChem.MolToSmiles(mol))
        for atom in copy_mol.GetAtoms():
            atom.SetAtomMapNum(0)
        return AllChem.MolToSmiles(copy_mol)

    # endregion

    # region ===== canonicalize map num =====

    @classmethod
    def canonical_rank_atoms(cls, mol):
        idx_to_old_map = {}
        for idx, atom in enumerate(mol.GetAtoms()):
            idx_to_old_map[idx] = atom.GetAtomMapNum()
            atom.SetAtomMapNum(0)
        atoms_ranks = list(AllChem.CanonicalRankAtoms(mol, breakTies=True))
        ranked_atom_idxes = []
        for r in range(len(atoms_ranks)):
            ranked_atom_idxes.append(atoms_ranks.index(r))
        return ranked_atom_idxes, idx_to_old_map

    @classmethod
    def canonicalize_map_num(cls, mol: Mol) -> Dict[int, int]:
        map_num_old_to_new = {}
        new_map_num = 1
        ranked_atom_idxes, idx_to_old_map = cls.canonical_rank_atoms(mol)
        for idx in ranked_atom_idxes:
            atom = mol.GetAtomWithIdx(idx)
            old_map_num = idx_to_old_map[idx]
            if old_map_num == 0:
                continue
            if old_map_num in map_num_old_to_new.keys():
                atom.SetAtomMapNum(map_num_old_to_new[old_map_num])
            else:
                atom.SetAtomMapNum(new_map_num)
                map_num_old_to_new[old_map_num] = new_map_num
                new_map_num += 1
        AllChem.CanonicalizeMol(mol)
        return map_num_old_to_new

    # endregion

    # region ===== edit mol =====

    @classmethod
    def merge_mols(cls, mols: [Mol]) -> Mol:
        res = None
        for n, mol in enumerate(mols):
            if n == 0:
                res = mol
            else:
                res = AllChem.CombineMols(res, mol)
        return res

    @classmethod
    def frag_idxes_to_mol(cls, mol: Mol, idxes: [int]):
        em = EditableMol(Mol())
        idxes_old_to_new = {}
        for i, idx in enumerate(idxes):
            em.AddAtom(mol.GetAtomWithIdx(idx))
            idxes_old_to_new[idx] = i

        for i, idx in enumerate(idxes):
            atom = mol.GetAtomWithIdx(idx)
            for bond in atom.GetBonds():
                if bond.GetBeginAtomIdx() == idx:
                    other_idx = bond.GetEndAtomIdx()
                else:
                    other_idx = bond.GetBeginAtomIdx()
                if other_idx < idx:
                    continue
                if other_idx in idxes_old_to_new.keys():
                    em.AddBond(idxes_old_to_new[idx], idxes_old_to_new[other_idx], bond.GetBondType())
        res = em.GetMol()
        res.ClearComputedProps()
        AllChem.GetSymmSSSR(res)
        res.UpdatePropertyCache(False)
        res._idxMap = idxes_old_to_new
        return res

    # endregion

    # region ===== query atom with map num =====

    @classmethod
    def contain_atom_with_map_num_in_mols(cls, mols: [Mol], map_num: int) -> bool:
        for _ in cls.query_atoms_with_map_num_in_mols(mols, map_num):
            return True
        return False

    @classmethod
    def query_atoms_with_map_num_in_mols(cls, mols: [Mol], map_num: int) -> Iterator[Atom]:
        for mol in mols:
            for atom in cls.query_atoms_with_map_num(mol, map_num):
                yield atom

    @classmethod
    def query_atoms_with_map_num(cls, mol: Mol, map_num: int) -> Iterator[Atom]:
        for atom in mol.GetAtoms():
            if atom.GetAtomMapNum() == map_num:
                yield atom

    # endregion


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/7/25 10:42
# @Author  : zhangbc0315@outlook.com
# @File    : rxn_center_detector.py
# @Software: PyCharm

from rdkit.Chem import AllChem
from rdkit.Chem.rdchem import Atom, Bond, Mol

from chem_utils.chem_handler import AtomHandler, MolHandler


class RxnCenterDetector:

    # region ===== assign special atoms =====

    @classmethod
    def _is_atom_dropped_by_product(cls, map_num: int, product: Mol) -> bool:
        return map_num is None or map_num == 0 \
               or not MolHandler.contain_atom_with_map_num_in_mols([product], map_num)

    @classmethod
    def _is_atom_dropped(cls, map_num: int, rxn: AllChem.ChemicalReaction) -> bool:
        return map_num is None or map_num == 0 \
               or not MolHandler.contain_atom_with_map_num_in_mols(rxn.GetProducts(), map_num)

    @classmethod
    def _has_same_num_neighbor(cls, r_atom: Atom, p_atom: Atom) -> bool:
        return len(r_atom.GetNeighbors()) == len(p_atom.GetNeighbors())

    @classmethod
    def _get_nbr_with_map_num(cls, atom: Atom, map_num: int) -> Atom:
        for nbr_atom in atom.GetNeighbors():
            if nbr_atom.GetAtomMapNum() == map_num:
                return nbr_atom
        return None

    @classmethod
    def _is_same_atoms(cls, atom1: Atom, atom2: Atom) -> bool:
        return atom1.GetAtomicNum() == atom2.GetAtomicNum() \
               and atom1.GetNumRadicalElectrons() == atom2.GetNumRadicalElectrons() \
               and len(atom1.GetNeighbors()) == len(atom2.GetNeighbors())

    @classmethod
    def _is_same_bond(cls, bond1: Bond, bond2: Bond):
        return bond1.GetBondType() == bond2.GetBondType()

    @classmethod
    def _has_same_neighbor_atom_with_same_bond(cls, r_atom: Atom, p_atom: Atom) -> bool:
        for r_nbr_atom in r_atom.GetNeighbors():
            p_nbr_atom = cls._get_nbr_with_map_num(p_atom, r_nbr_atom.GetAtomMapNum())
            if p_nbr_atom is None:
                return False
            r_bond = r_atom.GetOwningMol().GetBondBetweenAtoms(r_atom.GetIdx(), r_nbr_atom.GetIdx())
            p_bond = p_atom.GetOwningMol().GetBondBetweenAtoms(p_atom.GetIdx(), p_nbr_atom.GetIdx())
            if not cls._is_same_bond(r_bond, p_bond):
                return False
        return True

    @classmethod
    def _is_atom_reacted(cls, r_atom: Atom, p_atom: Atom) -> bool:
        if not cls._is_same_atoms(r_atom, p_atom):
            return True
        if not cls._has_same_neighbor_atom_with_same_bond(r_atom, p_atom):
            return True
        return False

    @classmethod
    def _assign_dropped_and_reacted_atoms_in_rxn(cls, rxn: AllChem.ChemicalReaction):
        for reactant in rxn.GetReactants():
            for atom in reactant.GetAtoms():
                map_num = atom.GetAtomMapNum()
                if cls._is_atom_dropped(map_num, rxn):
                    AtomHandler.set_atom_is_dropped(atom)
                    atom.SetAtomMapNum(0)
                else:
                    p_atom = list(MolHandler.query_atoms_with_map_num_in_mols(rxn.GetProducts(), map_num))[0]
                    if cls._is_atom_reacted(atom, p_atom):
                        AtomHandler.set_atom_is_reacted(atom)
                        AtomHandler.set_atom_is_reacted(p_atom)

    @classmethod
    def get_center_atom_idxes(cls, reactant: Mol, product: Mol):
        center_atom_idxes = []
        for atom in reactant.GetAtoms():
            map_num = atom.GetAtomMapNum()
            if cls._is_atom_dropped_by_product(map_num, product):
                center_atom_idxes.append(atom.GetIdx())
            else:
                p_atom = list(MolHandler.query_atoms_with_map_num(product, map_num))[0]
                if cls._is_atom_reacted(atom, p_atom):
                    center_atom_idxes.append(atom.GetIdx())
        return center_atom_idxes

    # endregion


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/7/19 15:08
# @Author  : zhangbc0315@outlook.com
# @File    : rxn_template_extractor.py
# @Software: PyCharm


class RxnTemplateExtractor:

    @classmethod
    def rxn_smiles_to_rt_smarts(cls, rxn_smiles: str):
        pass


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/8/22 10:48
# @Author  : zhangbc0315@outlook.com
# @File    : mol_smiles.py
# @Software: PyCharm
import copy

from rdkit.Chem.rdchem import Mol
from rdkit.Chem import AllChem


class MolSmiles:

    @classmethod
    def get_smiles_without_map_num(cls, rdmol: Mol) -> str:
        _rdmol = copy.deepcopy(rdmol)
        for atom in _rdmol.GetAtoms():
            atom.SetAtomMapNum(0)
        return AllChem.MolToSmiles(_rdmol)


if __name__ == "__main__":
    rxn = AllChem.MolFromSmiles("[F:1][C:2]1[CH:7]=[CH:6][C:5](Br)=[CH:4][N:3]=1")
    print(MolSmiles.get_smiles_without_map_num(rxn))



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/1/5 13:12
# @Author  : zhangbc0315@outlook.com
# @File    : atom_handler.py
# @Software: PyCharm

from rdkit.Chem.rdchem import Atom


class AtomHandler:

    TYPE_DROPPED = "dropped"
    TYPE_REACTED = "reacted"
    TYPE_NBR = "nbr"

    @classmethod
    def set_atom_is_nbr(cls, atom: Atom):
        atom.SetProp("type", cls.TYPE_NBR)

    @classmethod
    def is_atom_nbr(cls, atom: Atom) -> bool:
        return "type" in atom.GetPropNames() and atom.GetProp("type") == cls.TYPE_NBR

    @classmethod
    def set_atom_is_reacted(cls, atom: Atom):
        atom.SetProp("type", cls.TYPE_REACTED)

    @classmethod
    def is_atom_reacted(cls, atom: Atom) -> bool:
        return "type" in atom.GetPropNames() and atom.GetProp("type") == cls.TYPE_REACTED

    @classmethod
    def set_atom_is_dropped(cls, atom: Atom):
        atom.SetProp("type", cls.TYPE_DROPPED)

    @classmethod
    def is_atom_dropped(cls, atom: Atom) -> bool:
        return "type" in atom.GetPropNames() and atom.GetProp("type") == cls.TYPE_DROPPED


if __name__ == "__main__":
    pass



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2022/7/22 14:37
# @Author  : zhangbc0315@outlook.com
# @File    : rxn_checker.py
# @Software: PyCharm

from rdkit.Chem.AllChem import ChemicalReaction
from rdkit.Chem.rdchem import Mol


class RxnChecker:

    @classmethod
    def _get_all_map_nums_in_mols(cls, mols: [Mol]) -> [int]:
        map_nums = []
        for mol in mols:
            for atom in mol.GetAtoms():
                map_nums.append(atom.GetAtomMapNum())
        return map_nums

    @classmethod
    def is_mapped_enough(cls, rdrxn: ChemicalReaction):
        all_rs_map_nums = cls._get_all_map_nums_in_mols(rdrxn.GetReactants())
        for product in rdrxn.GetProducts():
            for p_atom in product.GetAtoms():
                if p_atom.GetAtomMapNum() == 0:
                    return False
                elif p_atom.GetAtomMapNum() not in all_rs_map_nums:
                    return False
        return True


if __name__ == "__main__":
    pass
